from dataclasses import dataclass

@dataclass
class JobResults:
    errors: int = 0

"""
Single-threaded job queue. This should be turned into a parallel job queue at some point.
"""
class JobQueue:
    def __init__(self, name=None, parent=None) -> None:
        self.name = name
        self.parent = parent
        self.waited = False
        self.subs = []
        self.logs = []
        self.results = parent.results if parent else JobResults()

    def __del__(self):
        if not self.waited and len(self.subs) > 0:
            print(f'X[{self._name()}]: Failed to wait for subordinate jobs!', flush=True)
        if self.logs:
            print(f'X[{self._name()}]: Failed to print logs!', flush=True)

    def submit(self, name, job):
        if name is None and self.name is not None:
            raise Exception('Queue must have a name if parent queue has a name')
        sub = JobQueue(name=name, parent=self)
        self.subs.append(sub)
        job(sub)
        sub.flush_logs()

    def flush_logs(self):
        for level, msg in self.logs:
            print(f'{level}[{self._name()}]: {msg}', flush=True)
        self.logs = []

    def error(self, msg):
        self.results.errors += 1
        self._log("E", msg)

    def info(self, msg):
        self._log("I", msg)

    def warning(self, msg):
        self._log("W", msg)

    def _log(self, level, msg):
        self.logs.append((level, msg))

    def wait(self):
        self.waited = True
        return self.results

    def is_root(self):
        return self.parent is None or self.name is None

    def _name(self):
        if self.is_root():
            return '(root)'
        else:
            return f'{self.parent._name()}/{self.name}'



from dataclasses import dataclass
from enum import Enum
import re

class MediaType(Enum):
    Video = 0,
    Audio = 1,
    Image = 2,

def compile_extension_regex(*extensions):
    return re.compile('|'.join([f'\\.{e}' for e in extensions]), flags=re.IGNORECASE)

FFMPEG_SUPPORTED_EXTENSIONS_BY_TYPE = {
    MediaType.Video: ["mp4", "m4v", "mov", "avi", "mkv", "mpeg", "mpg"],
    MediaType.Audio: ["aac", "mp3", "flac", "ogg", "m4a", "wav", "wma"],
    MediaType.Image: ["jpg", "jpeg", "png", "gif"]
}
FFMPEG_SUPPORTED_EXTENSIONS = sum(FFMPEG_SUPPORTED_EXTENSIONS_BY_TYPE.values(), [])
FFMPEG_SUPPORTED_EXTENSIONS_REGEX = {k: compile_extension_regex(*v) for k, v in FFMPEG_SUPPORTED_EXTENSIONS_BY_TYPE.items()} 

def ffmpeg_supports(filename):
    for type in FFMPEG_SUPPORTED_EXTENSIONS_BY_TYPE.keys():
        if ffmpeg_supports_type(type, filename):
            return True
    return False

def ffmpeg_supports_types(types, filename):
    for type in types:
        if ffmpeg_supports_type(type, filename):
            return True
    return False

def ffmpeg_supports_type(type, filename):
    return FFMPEG_SUPPORTED_EXTENSIONS_REGEX[type].match(filename.suffix)



from ntpath import realpath
import os

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from stat import S_ISDIR, S_ISREG, S_ISLNK
from typing import Callable, List

class EntryType(Enum):
    DIR = 0
    FILE_MEDIA = 1
    FILE_IGNORE = 2
    FILE_SPAM = 3

class SymlinkMode(Enum):
    Ignore = "ignore"
    Error = "error"
    Warn = "warn"
    Allow = "allow"
    AllowFile = "allowfile"
    AllowDir = "allowdir"

@dataclass
class PathScanResults:
    directory_list: List[Path]
    media_list: List[Path]
    unknown_extensions: List[str]

class PathScanner:
    def __init__(self,
        symlink_mode = SymlinkMode.Warn,
        supported_extension_matcher: Callable[[Path], bool] = None,
        ignored_pattern_matcher: Callable[[Path], bool] = None,
        spam_files_matcher: Callable[[Path], bool] = None) -> None:

        self.symlink_warned = False
        self.symlink_mode = symlink_mode
        self.supported_extension_matcher = supported_extension_matcher
        self.ignored_pattern_matcher = ignored_pattern_matcher
        self.spam_files_matcher = spam_files_matcher

    def scan(self, q, dir):
        files = []
        dirs = []
        unknown_extensions = set()
        for f in os.listdir(dir):
            filename = dir / f
            if self.ignored_pattern_matcher(filename):
                continue
            try:
                st = os.lstat(filename)
            except Exception as e:
                q.error(f"Unrecoverable filesystem error while trying to read {f} ({e})")
                continue

            if S_ISLNK(st.st_mode):
                if self.symlink_mode == SymlinkMode.Error:
                    q.error(f"Found unexpected symlink: {filename}")
                    continue
                elif self.symlink_mode == SymlinkMode.Warn:
                    if not self.symlink_warned:
                        q.warning(f"Ignoring symlinks, pass --symlinks=allow to allow this behavior")
                        self.symlink_warned = True
                    continue
                elif self.symlink_mode in [SymlinkMode.Allow, SymlinkMode.AllowDir, SymlinkMode.AllowFile]:
                    pass
                elif self.symlink_mode == SymlinkMode.Ignore:
                    continue
                else:
                    raise Exception(f"Unexpected symlink mode: {self.symlink_mode}")
                # Stat the underlying file or directory
                st = os.stat(filename)
                # Assume that the user only wanted the symlink type specified
                if S_ISREG(st.st_mode) and self.symlink_mode == SymlinkMode.AllowDir:
                    q.error(f"File symlink found and only directories are allowed: {filename}")
                    continue
                elif S_ISDIR(st.st_mode) and self.symlink_mode == SymlinkMode.AllowFile:
                    q.error(f"Directory symlink found and only files are allowed: {filename}")
                    continue

            # Ignore zero-length files
            if st.st_size == 0:
                continue
            if S_ISREG(st.st_mode):
                if self.supported_extension_matcher(filename):
                    files.append(filename)
                else:
                    if filename.suffix:
                        unknown_extensions.add(filename.suffix)
            elif S_ISDIR(st.st_mode):
                dirs.append(filename)
            pass
        unknown_extensions = list(unknown_extensions)
        unknown_extensions.sort()
        files.sort()
        dirs.sort()

        return PathScanResults(directory_list=dirs, media_list=files, unknown_extensions=unknown_extensions)



from pathlib import Path

class Docker:
    def __init__(self, pwd, prefixes) -> None:
        if pwd == None or prefixes == None:
            self.docker_prefix = Path('/')
            return

        # Test each candidate to see if we can find the pwd
        pwd = Path(pwd)
        prefixes = [Path(x) for x in prefixes]
        for candidate in prefixes:
            if (candidate / pwd.relative_to('/')).exists():
                self.docker_prefix = candidate
                return

        raise BaseException("Unable to locate docker root")

    def none():
        return Docker(None, None)

    def dockerize_path(self, path):
        return self.docker_prefix / Path(path).relative_to('/')



from dataclasses import dataclass
from pathlib import Path
from typing import List

from .ffmpeg import MediaType, ffmpeg_supports_types
from .forward_progress import subprocess_forward_progress
from .operation import Operation

@dataclass
class FFMPEGPreset:
    ext: str
    args: List[str]

FFMPEG_TRANSCODE_BASE_ARGS = [
    '-xerror',
    '-v', 'error', 
    '-y',
    '-i', '-']
FFMPEG_PRESETS = {
    'aac-64k': FFMPEGPreset(ext='m4a', args=['-vn', '-c:a', 'aac', '-b:a', '64k', '-f', 'mp4']),
    'aac-128k': FFMPEGPreset(ext='m4a', args=['-vn', '-c:a', 'aac', '-b:a', '128k', '-f', 'mp4']),
    'mp3-128k': FFMPEGPreset(ext='mp3', args=['-vn', '-c:a', 'mp3', '-b:a', '128k', '-f', 'mp3']),
    'mp3-320k': FFMPEGPreset(ext='mp3', args=['-vn', '-c:a', 'mp3', '-b:a', '320k', '-f', 'mp3']),
    'flac': FFMPEGPreset(ext='flac', args=['-vn', '-c:a', 'flac', '-f', 'flac']),
}

class FFMPEGTranscoderOperation(Operation):
    def __init__(self, output_dir: Path, transcode_args: List[str], extension: str) -> None:
        self.output_dir = output_dir
        self.transcode_args = FFMPEG_TRANSCODE_BASE_ARGS + transcode_args
        self.extension = f'.{extension}'
        
    def initialize(self, q, dir):
        q.info(f"Transcoding files: ffmpeg {' '.join(self.transcode_args)} [output-file] < [input-file]")
        self.root = dir
        self.output_dir.mkdir(parents=True, exist_ok=True)

    def operate(self, q, dir, files: List[Path]):
        for file in files:
            q.submit(file.name, lambda q: self._job(q, file))
        q.wait()

    def _job(self, q, file: Path):
        if ffmpeg_supports_types([MediaType.Video, MediaType.Audio], file):
            q.info(f"Transcoding...")
            out = self.output_dir / file.with_suffix(self.extension).relative_to(self.root)
            out.parent.mkdir(parents=True, exist_ok=True)
            args = self.transcode_args + [str(out)]
            errors = subprocess_forward_progress(file, args, "ffmpeg")
            if errors:
                q.error(errors)
            if out.exists():
                q.info(f"Size: {file.stat().st_size // 1024}k -> {out.stat().st_size // 1024}k")
            else:
                q.error("Failed to transcode file")



from .ffmpeg import ffmpeg_supports
from .jobqueue import JobQueue
from .forward_progress import subprocess_forward_progress
from .operation import Operation

FFMPEG_VERIFY_ARGS = [
        "-xerror",
        "-v", "error",
        "-i", "-",
        "-f", "null",
        "-"
    ]

def ffmpeg_validate(input, timeout=10, executable="ffmpeg", progress_callback=None):
    return subprocess_forward_progress(input, FFMPEG_VERIFY_ARGS, executable, timeout=timeout, progress_callback=progress_callback)

class FFMPEGValidateOperation(Operation):
    def initialize(self, q, dir):
        q.info(f"Verifying internal consistency of media files: ffmpeg {' '.join(FFMPEG_VERIFY_ARGS)} < [file] > /dev/null")

    def operate(self, q: JobQueue, dir, files):
        stats = { 'good': 0, 'bad': 0, 'ignored': 0 }
        for file in files:
            q.submit(file.name, lambda q: self._job(q, stats, file))
        q.wait()
        if stats['ignored']:
            q.info(f"{stats['good']} good file(s), {stats['bad']} bad file(s), {stats['ignored']} ignored file(s)")
        elif stats['bad']:
            q.info(f"{stats['good']} good file(s), {stats['bad']} bad file(s)")
        else:
            q.info(f"{stats['good']} good file(s)")

    def _job(self, q, stats, file):
        if ffmpeg_supports(file):
            errors = ffmpeg_validate(file)
            if errors:
                stats['bad'] += 1
                q.error(errors)
            else:
                stats['good'] += 1
        else:
            stats['ignored'] += 1

if __name__ == '__main__':
    import sys
    errors = ffmpeg_validate(sys.argv[1], timeout=1, executable="ffmpeg")
    if errors:
        print(errors)
    else:
        print("Ok!")
